/*
 * main.c:
 * Tunnel SLIP packets across a UDP link.
 *
 * Copyright (c) 2001 Chris Lightfoot. All rights reserved.
 * Email: chris@ex-parrot.com; WWW: http://www.ex-parrot.com/~chris/
 *
 */

static const char rcsid[] = "$Id: main.c,v 1.3 2001/08/31 20:41:11 chris Exp $";

#include <sys/types.h>

#include <errno.h>
#include <fcntl.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <syslog.h>
#include <time.h>
#include <unistd.h>

#include <netdb.h>
#include <arpa/inet.h>
#include <netinet/in.h>

#include <sys/ioctl.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/time.h>

#include "vpn-udp.h"

int verbose;

/* Flag to tell main loop to exit. */
sig_atomic_t foad;

/* write_pid_file:
 * Write our PID into a file; if the file exists, overwrite it if the PID in it
 * does not correspond to a real process. */
int write_pid_file(const char *file) {
    FILE *fp;
    struct stat st;

    if (stat(file, &st) == -1) {
        if (errno != ENOENT) {
            print_log(LOG_ERR, "stat(`%s'): %s", file, strerror(errno));
            return 0;
        }
    } else {
        pid_t pid;
        char buf[32];
        fp = fopen(file, "rt");
        if (!fp) {
            print_log(LOG_ERR, "fopen(`%s'): %s", file, strerror(errno));
            return 0;
        }
        fgets(buf, sizeof(buf), fp);
        pid = (pid_t)atoi(buf);
        if (pid <= 1)
            print_log(LOG_ERR, "%s: PID file exists, but is invalid; ignoring it", file);
        else if (kill(pid, 0) == -1)
            print_log(LOG_ERR, "%s: stale PID file", file);
        else {
            print_log(LOG_ERR, "%s: PID file refers to running process %d", file, (int)pid);
            return 0;
        }
        fclose(fp);
    }
    
    fp = fopen(file, "wt");
    if (!fp) {
        print_log(LOG_ERR, "fopen(`%s'): %s", file, strerror(errno));
        return 0;
    }
    fprintf(fp, "%d\n", (int)getpid());
    fclose(fp);
    if (verbose)
        print_log(LOG_ERR, "wrote PID to file `%s'", file);

    return 1;
}

/* delete_pid_file:
 * Remove the PID file. */
int delete_pid_file(const char *file) {
    return unlink(file) == 0;
}

/* terminate_signal_handler:
 * Signal handler to make main program quit. */
void terminate_signal_handler(int s) {
    foad = s;
}

/* setup_signals:
 * Set various signal handlers. */
void setup_signals() {
    int ignore_signals[]    = {SIGPIPE, SIGHUP, SIGALRM, SIGUSR1, SIGUSR2, SIGFPE, 0};
    int terminate_signals[] = {SIGINT, SIGTERM, SIGQUIT, SIGABRT, SIGSEGV, SIGBUS, 0};
    int *i;
    struct sigaction sa;
    memset(&sa, 0, sizeof(sa));

    for (i = ignore_signals; *i; ++i) {
        sa.sa_handler = SIG_IGN;
        sigaction(*i, &sa, NULL);
    }

    for (i = terminate_signals; *i; ++i) {
        sa.sa_handler = terminate_signal_handler;
        sigaction(*i, &sa, NULL);
    }
}

/* daemon:
 * Become a daemon. From `The Unix Programming FAQ', Andrew Gierth et al. */
int daemon(int nochdir, int noclose) {
    switch (fork()) {
        case 0:  break;
        case -1: return -1;
        default: _exit(0);          /* exit the original process */
    }

    if (setsid() < 0)               /* shouldn't fail */
        return -1;

    switch (fork()) {
        case 0:  break;
        case -1: return -1;
        default: _exit(0);
    }

    if (!nochdir) chdir("/");

    if (!noclose) {
        int i, j = sysconf(_SC_OPEN_MAX); /* getdtablesize()? */
        for (i = 0; i < j; ++i) close(i);
        open("/dev/null",O_RDWR);
        dup(0); dup(0);
    }

    return 0;
}

/* run_program:
 * Run a program, with the given argument; used to run the start/finish with
 * VPN programs. */
int run_program(const char *prog, const char *arg) {
    char *s = (char*)malloc(strlen(prog) + strlen(arg) + 2);
    int res;
    sprintf(s, "%s %s", prog, arg);
    errno = 0;
    /* FIXME don't use system but fork and copy stdout/stderr into syslog. */
    res = system(s);
    free(s);
    if ((res == 127 && errno != 0) || res == -1) {
        print_log(LOG_ERR, "system(`%s %s'): %s", prog, arg, strerror(errno));
        return 0;
    } else if (verbose)
        print_log(LOG_ERR, "%s %s: exited with code %d", prog, arg, res);

    return 1;
}

/* usage:
 * Print usage information. */
void usage(FILE *fp) {
    fprintf(fp,
"vpn-udp:\n"
"Create a host-host VPN (`tunnel') using UDP packets. This program will call\n"
"slattach(8) to configure a new SLIP interface associated with the local end\n"
"of the tunnel.\n"
"\n"
"Synopsis: vpn-udp [options] keyfile port host[:port]\n"
"\n"
"Options:\n"
"   -h              Display this message.\n"
"   -v              Be verbose.\n"
"   -d              Do not detach from controlling terminal.\n"
"   -e              Print diagnostics to stderr as well as logging them to the\n"
"                   system log.\n"
"   -l interval     Use interval as the link latency margin\n"
"   -p file         Write PID to file.\n"
"   -s program      Run program when the SLIP interface is created; the name of\n"
"                   the network interface, for instance `sl0', will be appended\n"
"                   to its command line.\n"
"   -f program      Run program when the tunnel is shut down.\n"
"\n"
"Keyfile is a file from which 16 bytes of cryptographic key data are read.\n"
"This must be the same on both ends of the link. The first port is the local\n"
"port on which to listen. Host specifies the real internet address of the\n"
"other end of the tunnel. The second port is the port on the remote host to\n"
"which tunnel packets should be sent. If not specified, it is inferred from\n"
"the origin of the first packet received at this end of the tunnel. This\n"
"feature can only be used on one end of the link.\n");
}

/* main:
 * Entry point. */
char optstr[] = "+hvdel:p:s:f:";
int log_stderr;

int main(int argc, char *argv[]) {
    int c;
    char *keyfile, *lport, *raddr, *pidfile = NULL;
    char *start_program = NULL, *finish_program = NULL;
    char *p;
    struct sockaddr_in sin = {0};
    int ret = -1, timeout = 0;
    time_t latency_margin = 16;
    int nodetach;
    
    extern time_t last_sent, last_received; /* in tunnel.c */
    extern int fd_tunnel, fd_if;            /* in tunnel.c and if.c */
    extern struct in_addr remote_addr;      /* in tunnel.c */
    extern int local_port, remote_port;     /* in tunnel.c */
    extern char *if_name;                   /* in if.c */

    last_sent = last_received = time(NULL);

    /* Figure out the local and remote addresses to use. */
    opterr = 0;
    while ((c = getopt(argc, argv, optstr)) != -1) {
        switch (c) {
            case 'h':
                usage(stdout);
                return 0;

            case 'v':
                verbose = 1;
                break;

            case 'd':
                nodetach = 1;
                break;
                
            case 'e':
                log_stderr = 1;
                break;

            case 'l':
                latency_margin = (time_t)atoi(optarg);
                if (!latency_margin || latency_margin < 4) {
                    fprintf(stderr, PROGNAME": `%s' is not a valid latency margin\n", optarg);
                    return -1;
                }
                break;

            case 'p':
                pidfile = optarg;

            case 's':
                start_program = optarg;
                break;

            case 'f':
                finish_program = optarg;
                break;

            case '?':
            default:
                if (strchr(optstr, optopt))
                    fprintf(stderr, PROGNAME": option -%c requires an argument\n", optopt);
                else
                    fprintf(stderr, PROGNAME": unknown option -%c\n", optopt);
                usage(stderr);
                return -1;
        }
    }

    /* Verify that we have the requisite three arguments... */
    if (optind + 2 > argc) {
        fprintf(stderr, PROGNAME": not enough arguments\n");
        usage(stderr);
        return -1;
    }

    /* ... and that we have sufficient privileges. */
    if (getuid() != 0) {
        fprintf(stderr, PROGNAME": must be run as root\n");
        return -1;
    }

    setup_signals();

    /* Become a daemon, unless told otherwise. */
    if (!nodetach)
        daemon(0, 0);

    /* Start system logging. */
    openlog(PROGNAME, LOG_PID | LOG_NDELAY, LOG_DAEMON);
    
    keyfile = argv[optind];
    lport   = argv[optind + 1];
    raddr   = argv[optind + 2];

    /* Get key data and set up Blowfish key. */
    if (!setup_encryption(keyfile)) return -1;

    /* Figure out local and remote addresses. */
    local_port = (short)atoi(lport);
    if (!local_port) {
        print_log(LOG_ERR, "`%s' is not a valid local port", lport);
        return -1;
    }

    p = strchr(raddr, ':');
    if (p) {
        *p++ = 0;
        remote_port = (short)atoi(p);
        if (!remote_port) {
            print_log(LOG_ERR, "`%s' is not a valid remote port", p);
            return -1;
        }
    }

    if (!inet_aton(raddr, &remote_addr)) {
        struct hostent *he;
        he = gethostbyname(raddr);
        if (!he) {
            print_log(LOG_ERR, "host `%s' cannot be found", raddr);
            return -1;
        } else memcpy(&remote_addr, he->h_addr, sizeof(struct in_addr));
    }

    /* Obtain a source of random data for Cryptographic Purposes(TM). */
    if (!setup_random_data()) return -1;
    
    /* Save our PID. */
    if (pidfile && !write_pid_file(pidfile)) return -1;
    
    /* Set up the SLIP interface. */
    if (!create_interface()) return -1;

    if (start_program && !run_program(start_program, if_name)) goto finish;

    /* Set up the local UDP socket. */
    fd_tunnel = socket(AF_INET, SOCK_DGRAM, 0);
    if (fd_tunnel == -1) {
        print_log(LOG_ERR, "socket: %m");
        goto finish;
    }

    /* Bind socket to local port. */
    sin.sin_family = AF_INET;
    sin.sin_port = htons((short)local_port);
    if (bind(fd_tunnel, &sin, sizeof(sin)) == -1) {
        print_log(LOG_ERR, "bind: %m");
        goto finish;
    } else if (verbose)
        print_log(LOG_ERR, "listening on local port %d", local_port);

    if (remote_port && !connect_to_remote()) goto finish;

    /* Now, we need to await packets arriving either on the local SLIP
     * interface, which should be relayed to the remote host; or from the
     * network, which should be fed into the local SLIP interface. */
    do {
        fd_set fds;
        struct timeval tv = {0, 100000};
        int n = fd_if;
        FD_ZERO(&fds);
        if (remote_port) FD_SET(fd_if, &fds);
        FD_SET(fd_tunnel, &fds);
        if (fd_tunnel > n) n = fd_tunnel;
        switch (select(n + 1, &fds, NULL, NULL, &tv)) {
            case -1:
                if (errno != EINTR) {
                    print_log(LOG_ERR, "select: %m");
                    goto finish;
                }
                break;

            case 0:
                break;

            default:
                if (FD_ISSET(fd_if, &fds) && !read_from_if())
                    goto finish;
                if (FD_ISSET(fd_tunnel, &fds) && !read_from_tunnel())
                    goto finish;
                break;
        }

        if (last_received + latency_margin < time(NULL)) {
            /* We haven't heard anything from the other end for a while; it's
             * probably dead. Quit, and let the calling program sort things
             * out. */
            print_log(LOG_ERR, "no packets received from network for %d seconds; quitting", (int)(time(NULL) - last_received));
            timeout = 1;
        }

        if (last_sent + latency_margin / 4 < time(NULL)) {
            /* We haven't sent anything for quite a while; send a NOP so that
             * the other end doesn't think we've died. */
            if (verbose) print_log(LOG_ERR, "sending keepalive after %d seconds of inactivity", (int)(time(NULL) - last_sent));
            if (!write_to_tunnel(NULL, 0)) timeout = 1;
        }
    } while (!foad && !timeout);

    if (foad) print_log(LOG_ERR, "quitting on signal %d", (int)foad);
    ret = 0;

finish:
    if (finish_program) run_program(finish_program, if_name);
    if (pidfile) delete_pid_file(pidfile);

    print_log(LOG_INFO, "exiting");

    return ret;
}
